from enum import Enum

__all__ = ["EndpointNames"]


class EndpointNames(Enum):
    """Standardised enum of names used in JSON schema filenames for each endpoint."""

    ACCIDENT_STATS = "AccidentStats"
    AIR_QUALITY = "AirQuality"
    BIKE_POINT = "BikePoint"
    CROWDING = "crowding"
    DISRUPTIONS_LIFTS_V2 = "Disruptions-Lifts-v2"
    JOURNEY = "Journey"
    LINE = "Line"
    MODE = "Mode"
    OCCUPANCY = "occupancy"
    PLACE = "Place"
    ROAD = "Road"
    SEARCH = "Search"
    STOP_POINT = "StopPoint"
    VEHICLE = "Vehicle"



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class RoadCorridor(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.RoadCorridor
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    DisplayName: str = None
    Group: str = None
    StatusSeverity: str = None
    StatusSeverityDescription: str = None
    Bounds: str = None
    Envelope: str = None
    StatusAggregationStartDate: datetime = None
    StatusAggregationEndDate: datetime = None
    Url: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl')


RoadCorridorModel = RoadCorridor


class DbGeographyWellKnownValue(BaseModel):
    """
    Autogenerated from Road::System.Data.Spatial.DbGeographyWellKnownValue
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    CoordinateSystemId: int = None
    WellKnownText: str = None
    WellKnownBinary: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='System')


DbGeographyWellKnownValueModel = DbGeographyWellKnownValue


class DbGeography(BaseModel):
    """
    Autogenerated from Road::System.Data.Spatial.DbGeography
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Geography: DbGeographyWellKnownValueModel = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='System-2')


DbGeographyModel = DbGeography


class StreetSegment(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.StreetSegment
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Toid: str = None
    LineString: str = None
    SourceSystemId: int = None
    SourceSystemKey: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-2')


StreetSegmentModel = StreetSegment


class Street(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.Street
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Name: str = None
    Closure: str = None
    Directions: str = None
    Segments: list["StreetSegmentModel"]
    SourceSystemId: int = None
    SourceSystemKey: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-3')


StreetModel = Street


class RoadProject(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.RoadProject
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    ProjectId: str = None
    SchemeName: str = None
    ProjectName: str = None
    ProjectDescription: str = None
    ProjectPageUrl: str = None
    ConsultationPageUrl: str = None
    ConsultationStartDate: datetime = None
    ConsultationEndDate: datetime = None
    ConstructionStartDate: datetime = None
    ConstructionEndDate: datetime = None
    BoroughsBenefited: list[str]
    CycleSuperhighwayId: str = None
    Phase: str = None
    ContactName: str = None
    ContactEmail: str = None
    ExternalPageUrl: str = None
    ProjectSummaryPageUrl: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-4')


RoadProjectModel = RoadProject


class RoadDisruptionLine(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.RoadDisruptionLine
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: int = None
    RoadDisruptionId: str = None
    IsDiversion: bool = None
    MultiLineString: DbGeographyModel = None
    StartDate: datetime = None
    EndDate: datetime = None
    StartTime: str = None
    EndTime: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-5')


RoadDisruptionLineModel = RoadDisruptionLine


class RoadDisruptionImpactArea(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.RoadDisruptionImpactArea
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: int = None
    RoadDisruptionId: str = None
    Polygon: DbGeographyModel = None
    StartDate: datetime = None
    EndDate: datetime = None
    StartTime: str = None
    EndTime: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-6')


RoadDisruptionImpactAreaModel = RoadDisruptionImpactArea


class RoadDisruptionSchedule(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.RoadDisruptionSchedule
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    StartTime: datetime = None
    EndTime: datetime = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-7')


RoadDisruptionScheduleModel = RoadDisruptionSchedule


class RoadDisruption(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.RoadDisruption
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Url: str = None
    Point: str = None
    Severity: str = None
    Ordinal: int = None
    Category: str = None
    SubCategory: str = None
    Comments: str = None
    CurrentUpdate: str = None
    CurrentUpdateDateTime: datetime = None
    CorridorIds: list[str]
    StartDateTime: datetime = None
    EndDateTime: datetime = None
    LastModifiedTime: datetime = None
    LevelOfInterest: str = None
    Location: str = None
    Status: str = None
    Geography: DbGeographyModel = None
    Geometry: DbGeographyModel = None
    Streets: list["StreetModel"]
    IsProvisional: bool = None
    HasClosures: bool = None
    LinkText: str = None
    LinkUrl: str = None
    RoadProject: RoadProjectModel = None
    PublishStartDate: datetime = None
    PublishEndDate: datetime = None
    TimeFrame: str = None
    RoadDisruptionLines: list["RoadDisruptionLineModel"]
    RoadDisruptionImpactAreas: list["RoadDisruptionImpactAreaModel"]
    RecurringSchedules: list["RoadDisruptionScheduleModel"]
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-8')


RoadDisruptionModel = RoadDisruption


class Object(BaseModel):
    """
    Autogenerated from Road::System.Object
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='System-3')


ObjectModel = Object


class StatusSeverity(BaseModel):
    """
    Autogenerated from Road::Tfl.Api.Presentation.Entities.StatusSeverity
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    ModeName: str = None
    SeverityLevel: int = None
    Description: str = None
    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Tfl-9')


StatusSeverityModel = StatusSeverity


class Get200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Road::Get200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='Get200ApplicationJsonResponse')


Get200ApplicationJsonResponseModel = Get200ApplicationJsonResponse


class idsGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Road::ids-Get200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='ids-Get200ApplicationJsonResponse')


idsGet200ApplicationJsonResponseModel = idsGet200ApplicationJsonResponse


class idsStatusGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Road::ids-StatusGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='ids-StatusGet200ApplicationJsonResponse')


idsStatusGet200ApplicationJsonResponseModel = idsStatusGet200ApplicationJsonResponse


class idsDisruptionGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Road::ids-DisruptionGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='ids-DisruptionGet200ApplicationJsonResponse')


idsDisruptionGet200ApplicationJsonResponseModel = idsDisruptionGet200ApplicationJsonResponse


class MetaCategories(BaseModel):
    """
    Autogenerated from Road::MetaCategoriesGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='MetaCategoriesGet200ApplicationJsonResponse')


MetaCategoriesModel = MetaCategories


class MetaSeveritiesGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Road::MetaSeveritiesGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Road')
    _component_schema_name: str = PrivateAttr(default='MetaSeveritiesGet200ApplicationJsonResponse')


MetaSeveritiesGet200ApplicationJsonResponseModel = MetaSeveritiesGet200ApplicationJsonResponse


class Deserialisers(DtoEnum):
    Tfl = RoadCorridor
    System = DbGeographyWellKnownValue
    System_2 = DbGeography
    Tfl_2 = StreetSegment
    Tfl_3 = Street
    Tfl_4 = RoadProject
    Tfl_5 = RoadDisruptionLine
    Tfl_6 = RoadDisruptionImpactArea
    Tfl_7 = RoadDisruptionSchedule
    Tfl_8 = RoadDisruption
    System_3 = Object
    Tfl_9 = StatusSeverity
    Get200ApplicationJsonResponse = Get200ApplicationJsonResponse
    ids_Get200ApplicationJsonResponse = idsGet200ApplicationJsonResponse
    ids_StatusGet200ApplicationJsonResponse = idsStatusGet200ApplicationJsonResponse
    ids_DisruptionGet200ApplicationJsonResponse = idsDisruptionGet200ApplicationJsonResponse
    MetaCategoriesGet200ApplicationJsonResponse = MetaCategories
    MetaSeveritiesGet200ApplicationJsonResponse = MetaSeveritiesGet200ApplicationJsonResponse


from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class Object(BaseModel):
    """
    Autogenerated from AirQuality::System.Object
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='AirQuality')
    _component_schema_name: str = PrivateAttr(default='System.Object')


ObjectModel = Object


class Deserialisers(DtoEnum):
    System__Object = Object


from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import orjson

__all__ = ["ApiSchema"]


@dataclass
class ApiSchema:
    path: Path
    unified: bool

    def __post_init__(self):
        if self.path.suffix != ".json":
            raise ValueError(f"{self.path=} is not a JSON file")
        obj = orjson.loads(self.path.read_text())
        # Discard any of the entities with 'Array' in, they're not useful
        self.entity_schemas = {}
        self.ents_by_property_count = {}
        # self.entity_property_counts_inv = {}
        for name, schema in obj["components"].get("schemas", {}).items():
            # if self.is_skipped_name(name=name):
            #     continue
            self.entity_schemas[name] = schema
            property_count = len(schema.get("properties", []))
            # self.entity_property_counts[name] = property_count
            self.ents_by_property_count.setdefault(property_count, [])
            self.ents_by_property_count[property_count].append(name)

    @property
    def name(self) -> str:
        return self.path.stem

    # def is_skipped_name(self, name: str) -> bool:
    #    return "Array" in name



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class Bay(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.Bay
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    BayType: str = None
    BayCount: int = None
    Free: int = None
    Occupied: int = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl')


BayModel = Bay


class CarParkOccupancy(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.CarParkOccupancy
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Bays: list["BayModel"]
    Name: str = None
    CarParkDetailsUrl: str = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl-2')


CarParkOccupancyModel = CarParkOccupancy


class ChargeConnectorOccupancy(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.ChargeConnectorOccupancy
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: int = None
    SourceSystemPlaceId: str = None
    Status: str = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl-3')


ChargeConnectorOccupancyModel = ChargeConnectorOccupancy


class BikePointOccupancy(BaseModel):
    """
    Autogenerated from occupancy::Tfl.Api.Presentation.Entities.BikePointOccupancy
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Name: str = None
    BikesCount: int = None
    EmptyDocks: int = None
    TotalDocks: int = None
    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='Tfl-4')


BikePointOccupancyModel = BikePointOccupancy


class CarParkGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from occupancy::CarParkGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='CarParkGet200ApplicationJsonResponse')


CarParkGet200ApplicationJsonResponseModel = CarParkGet200ApplicationJsonResponse


class ChargeConnectoridsGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from occupancy::ChargeConnector-ids-Get200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='ChargeConnector-ids-Get200ApplicationJsonResponse')


ChargeConnectoridsGet200ApplicationJsonResponseModel = ChargeConnectoridsGet200ApplicationJsonResponse


class ChargeConnectorGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from occupancy::ChargeConnectorGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='ChargeConnectorGet200ApplicationJsonResponse')


ChargeConnectorGet200ApplicationJsonResponseModel = ChargeConnectorGet200ApplicationJsonResponse


class BikePointsidsGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from occupancy::BikePoints-ids-Get200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='occupancy')
    _component_schema_name: str = PrivateAttr(default='BikePoints-ids-Get200ApplicationJsonResponse')


BikePointsidsGet200ApplicationJsonResponseModel = BikePointsidsGet200ApplicationJsonResponse


class Deserialisers(DtoEnum):
    Tfl = Bay
    Tfl_2 = CarParkOccupancy
    Tfl_3 = ChargeConnectorOccupancy
    Tfl_4 = BikePointOccupancy
    CarParkGet200ApplicationJsonResponse = CarParkGet200ApplicationJsonResponse
    ChargeConnector_ids_Get200ApplicationJsonResponse = ChargeConnectoridsGet200ApplicationJsonResponse
    ChargeConnectorGet200ApplicationJsonResponse = ChargeConnectorGet200ApplicationJsonResponse
    BikePoints_ids_Get200ApplicationJsonResponse = BikePointsidsGet200ApplicationJsonResponse


from .types import RouteEnum


class VehicleEndpointRoutes(RouteEnum):
    EMISSION_SURCHARGE = "/EmissionSurcharge"
    """Gets the Emissions Surcharge compliance for the Vehicle"""
    ULEZ_COMPLIANCE = "/UlezCompliance"
    """Gets the Ulez Surcharge compliance for the Vehicle"""
    VEHICLE_ARRIVALS = "/{ids}/Arrivals"
    """Gets the predictions for a given list of vehicle Id's."""



import ast
from itertools import starmap


def python_type(json_type: str, format: str = None) -> str:
    """Map property types in the JSON schema to Python types for type annotation."""
    if json_type == "string":
        if format == "email":
            return "str"
        return "str"
    elif json_type == "integer":
        return "int"
    elif json_type == "boolean":
        return "bool"
    elif json_type == "array":
        return "list"
    elif json_type == "object":
        return "dict"
    else:
        return "Any"


def import_node(module: str, names: list[str]) -> ast.Import:
    if names:
        return ast.ImportFrom(
            module=module,
            names=[ast.alias(name=name) for name in names],
            level=0,
        )
    else:
        return ast.Import(names=[ast.alias(name=module)])


def generate_dataclass(schema: dict) -> str:
    """Generate a dataclass from a schema, importing the `field` function if there are any
    array properties requiring it.
    """
    import_list = {
        "json": [],
        "dataclasses": ["dataclass"],
        "pathlib": ["Path"],
        "dataclass_wizard": ["JSONWizard", "LoadMeta"],
        "dataclass_wizard.loaders": ["fromdict"],
        "jsonschema": [],
    }
    class_name = schema.get("title", "UntitledDataclass").replace(" ", "")
    properties = schema.get("properties", {})
    required = schema.get("required", [])
    contains_list = False
    dc_source = "@dataclass\n"
    dc_source += f"class {class_name}(JSONWizard):\n"
    for prop_name, prop_schema in properties.items():
        prop_type = python_type(prop_schema["type"], prop_schema.get("format"))
        is_array_prop = prop_schema["type"] == "array"
        if is_array_prop and "items" in prop_schema:
            item_type = python_type(prop_schema["items"]["type"])
            prop_type = f"{prop_type}[{item_type}]"
        if prop_name in required:
            default = ""
        else:
            if is_array_prop:
                contains_list = True
                default = " = field(default_factory=list)"
            else:
                default = " = None"
        dc_source += f"    {prop_name}: {prop_type}{default}\n"
    dc_source += """    @classmethod
    def from_dict(cls, o):
        jsonschema.validate(o, schema)
        return fromdict(cls, o)"""
    if contains_list:
        import_list["dataclasses"].append("field")
    imports = "\n".join(map(ast.unparse, starmap(import_node, import_list.items())))
    meta_binding = f"LoadMeta(raise_on_unknown_json_key=True).bind_to({class_name})"
    output = "\n\n".join([imports, dc_source, meta_binding])
    return output


schema = {
    "title": "Example Schema",
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "age": {"type": "integer"},
        "email": {"type": "string", "format": "email"},
        "friends": {"type": "array", "items": {"type": "string"}},
    },
    "required": ["name", "age"],
}

generated_code = generate_dataclass(schema)
print(generated_code)



from __future__ import annotations

import subprocess
from contextlib import AbstractContextManager
from dataclasses import dataclass

from ..utils.logs import set_up_logging
from ..utils.paths import db_path

__all__ = [
    "start_mongod",
    "MongodProcess",
    "get_mongod_process",
    "MongodExceptionGuard",
]

logger = set_up_logging(__name__)


def start_mongod() -> None:
    db_path.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Creating a mongod process at {db_path}")
    subprocess.Popen(["mongod", "--dbpath", db_path], stdout=subprocess.DEVNULL)
    logger.debug(f"Created a mongod process at {db_path}")


@dataclass
class MongodProcess:
    pid: str  # Numeric but still string
    fresh: bool

    def kill(self):
        subprocess.run(["kill", self.pid])

    def clean_up(self):
        if self.fresh:
            self.kill()


def get_mongod_process() -> MongodProcess:
    """Create new ``mongod`` process if one is not found. Either way, return the pid of the
    running process. Also return a bool to indicate whether this was created afresh,
    so that the caller can decide whether to kill the process when finished.
    """
    proc = subprocess.run(["pgrep", "mongod"], capture_output=True)
    nonexisting = proc.returncode != 0
    if nonexisting:
        start_mongod()
        proc = subprocess.run(["pgrep", "mongod"], capture_output=True)
        # If we still couldn't create a mongod process, give up?
        proc.check_returncode()
    pid = proc.stdout.decode().splitlines()[0]
    return MongodProcess(pid=pid, fresh=nonexisting)


# Idea via https://stackoverflow.com/a/60367619/
class MongodExceptionGuard(AbstractContextManager):
    def __init__(self, mongod_proc=get_mongod_process()):
        """Set up the Mongod daemon if a process is not found."""
        self.mongod_proc = mongod_proc

    def __exit__(self, exc_type, exc_value, traceback):
        """If the Mongod daemon was started 'fresh' upon initialisation, kill the process
        upon exitting the context manager (before raising any errors if present).
        """
        if exc_type:
            logger.error(f"{exc_type}, {exc_value}, {traceback}")
            self.mongod_proc.clean_up()
            raise
        else:
            self.mongod_proc.clean_up()



import logging
from textwrap import indent

__all__ = ["logger"]

logger = logging.getLogger("tubeulator.codegen.gen")


class CustomFormatter(logging.Formatter):
    def formatException(self, exc_info):
        result = super().formatException(exc_info)
        return indent(
            f"\n{result}\n",
            prefix=" " * 8,
        )


formatter = CustomFormatter("[%(levelname)s] - %(message)s")

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
console_handler.setFormatter(formatter)

logger.addHandler(console_handler)



from pathlib import Path

from .. import generated
from ..openapi.scan import count_namespace
from .gen import emit_deserialisers

__all__ = ["generate_schema_coverage"]

codegen_dir = Path(generated.__file__).parent

schema_names = sorted(count_namespace(ignore_responses=True), key=str.lower)
schema_module_names = [f"{schema}.py".replace("-", "") for schema in schema_names]


def generate_schema_coverage():
    for schema_module, schema_name in zip(schema_module_names, schema_names):
        module_code = emit_deserialisers(schema_name=schema_name)
        module_filepath = codegen_dir / schema_module
        module_filepath.write_text(module_code)



import ast
import textwrap
from itertools import starmap
from typing import Literal

from ...openapi.scan import scan_namespace
from ...utils.paths import SchemaPath, load_endpoint_component_schemas
from ...utils.string_conv import to_pascal_case
from .jsonschema import python_type
from .resolution import (
    find_array_literals,
    find_arrays,
    find_backrefs,
    find_chased,
    find_string_literals,
)

__all__ = [
    "import_node",
    "generate_dataclass_name",
    "generate_dataclass",
    "hidden_field",
    "generate_source",
]


def import_node(module: str, names: list[str]) -> ast.Import:
    if names:
        return ast.ImportFrom(module=module, names=[*map(ast.alias, names)], level=0)
    else:
        return ast.Import(names=[ast.alias(name=module)])


def generate_dataclass_name(
    schema: dict,
    name: str,
    source_schema_name: str,
    idx: int = None,
    dealiased_name: str | None = None,
    indent_level: int = 1,
) -> str | None:
    name, _ = generate_dataclass(
        schema=schema,
        name=name,
        source_schema_name=source_schema_name,
        idx=idx,
        dealiased_name=dealiased_name,
        indent_level=indent_level,
        schema_mapping=None,
        name_only=True,
    )
    return name


def generate_dataclass(
    schema: dict,
    name: str,
    source_schema_name: str,
    idx: int = None,
    dealiased_name: str | None = None,
    indent_level: int = 1,
    schema_mapping: dict[str, str] | None = None,
    name_only: bool = False,
) -> tuple[str, str] | tuple[None, None]:
    """Generate a dataclass from a schema, importing the `field` function if there are any
    array properties requiring it. Use the `idx` to indicate if we're generating
    multiple dataclasses in a loop. (TODO: replace with `schema: dict or list[dict]`).

    Either returns the name and the source, or None twice.
    """
    parent_schema = load_endpoint_component_schemas(source_schema_name)
    # ref may be None indicating schema may be {'type': 'object'}
    ref = schema.get("items", {}).get("$ref")
    ref_name = None if ref is None else SchemaPath(schema).ref.name
    backrefs = find_backrefs(parent_schema)
    schema_arrays = find_array_literals(parent_schema)
    schema_array_shortlist = find_arrays(list(schema_arrays))
    schema_strings = find_string_literals(parent_schema)
    schema_string_shortlist = find_arrays(list(schema_strings))
    # For Search, we also need to match Response DTOs without refs
    # For this we need to look for any arrays, not just $ref items
    # Chase the references if needed
    chased = find_chased(backrefs, name=name, ref_name=ref_name)
    app_infix = "Application"
    json_suffix = "JsonResponse"
    aj_suffix = f"{app_infix}{json_suffix}"
    response_shortlist = [c for c in chased if aj_suffix in c]
    if any("Array" in c for c in chased):
        array_shortlist = find_arrays(chased)
    else:
        array_shortlist = []
    assert not (response_shortlist and array_shortlist), "Got both Response and Array"
    if chased:
        # Multiple map to the node: they will be different response types
        assert response_shortlist or array_shortlist, "Neither Response nor Array found"
        if response_shortlist:
            if name not in response_shortlist:
                return None, None
            chase_prefixes = [
                p[: -len(aj_suffix)] for p in chased if p.endswith(aj_suffix)
            ]
            # old_chase_prefix = "_or_".join(chase_prefixes)  # Hotfix
            if len(chase_prefixes) > 1:
                chase_prefix = chase_prefixes[response_shortlist.index(name)]
            else:
                chase_prefix = chase_prefixes[0] if chase_prefixes else ""
            class_name = f"{chase_prefix}{aj_suffix}"
            gen_source = name  # ref_name
        else:
            if name not in array_shortlist:
                # Don't bother generating source for non-shortlisted array literals
                # it would only give None anyway
                return None, None
            ar_suffix = "Array"
            chase_prefixes = [
                p[: p.index(ar_suffix)] for p in chased if p in array_shortlist
            ]
            chase_prefix = "_or_".join(chase_prefixes)  # Hotfix: never used
            class_name = f"{chase_prefix}{ar_suffix}"
            gen_source = name  # ref_name
    elif dealiased_name is not None:
        stem = dealiased_name.rsplit(".", 1)[-1]
        class_name = f"{stem}"
        gen_source = dealiased_name
    elif (name, ref) != (None, None):
        # If they have a name but it's not in namespace, they're a response (targeting $ref)
        ns = scan_namespace(ignore_responses=True)
        if ref_name:
            if ref_lookup := ns[source_schema_name].get(ref_name, [None]):
                dealiased_ref = (ref_lookup)[0] or source_schema_name
            else:
                # The list of namespace lookups is empty
                dealiased_ref = source_schema_name
            # referent_class_name = f'{dealiased_ref.rsplit(".", 1)[-1]}'
            class_name = f'{dealiased_ref.rsplit(".", 1)[-1]}Response'
            gen_source = ref_name
        elif name in schema_arrays:
            if name not in schema_array_shortlist:
                return None, None
            short_delim = "Get200"
            shortname = name[: name.find(short_delim)] if short_delim in name else name
            class_name = f"{shortname}"
            gen_source = name
        else:
            if name in schema_strings and name not in schema_string_shortlist:
                return None, None
            # Essentially 'Unknown' f"UNK_{name}"
            class_name = f"{name}"
            gen_source = name
    else:
        gen_source = f"{source_schema_name}:{idx}"
        class_name = f"{source_schema_name}{idx if idx is not None else ''}"
        # class_name = source_schema_name.replace(" ", "")
    ent_prefix = "TflApiPresentationEntities"
    # Unslug the class name
    preproc_class_name = class_name.replace("-", "")
    # Remove the long-winded entity prefix from the class name
    preproc_class_name = preproc_class_name[
        len(ent_prefix) if preproc_class_name.startswith(ent_prefix) else 0 :
    ]
    if name_only:
        output = None
    else:
        output = generate_source(
            class_name=preproc_class_name,
            component_name=name,
            schema_name=source_schema_name,
            gen_source=gen_source,
            schema=schema,
            parent_schema=parent_schema,
            indent_level=indent_level,
            idx=idx,
            schema_mapping=schema_mapping,
        )
    return preproc_class_name, output


def hidden_field(default_value: str, style: Literal["pydantic", "jsonw"]) -> str:
    if style == "jsonw":
        value = f"field(default={default_value!r}, repr=False)"
    elif style == "pydantic":
        value = f"PrivateAttr(default={default_value!r})"
    return value


def make_class_header(class_name: str, style: Literal["pydantic", "jsonw"]) -> str:
    if style == "jsonw":
        hed = f"@dataclass\nclass {class_name}(JSONWizard):\n"
    elif style == "pydantic":
        hed = f"class {class_name}(BaseModel):\n"
    return hed


def generate_source(
    class_name: str,
    component_name: str,
    schema_name: str,
    gen_source: str,
    schema: dict,
    parent_schema: dict,
    indent_level: int,
    idx: int,
    schema_mapping: dict[str, str] | None,
    style: Literal["pydantic", "jsonw"] = "pydantic",
) -> str:
    is_jsonw = style == "jsonw"
    is_pydantic = style == "pydantic"
    properties = schema.get("properties", {})
    required = schema.get("required", [])
    contains_list = False
    dc_source = make_class_header(class_name=class_name, style=style)
    docstring = f"Autogenerated from {schema_name}::{gen_source}"
    dc_source += textwrap.indent(f'"""\n{docstring}\n"""\n', " " * indent_level * 4)
    if is_pydantic:
        alias_gen = "alias_generator=AliasGenerator(validation_alias=to_camel_case),"
        model_config = f"model_config = ConfigDict(\n    {alias_gen}\n)"
        dc_source += textwrap.indent(f"\n{model_config}\n\n", " " * indent_level * 4)
    for prop_name, prop_schema in properties.items():
        is_substituted = (
            "type" not in prop_schema and prop_schema and "$ref" in prop_schema
        )
        if is_substituted:
            referent = prop_schema["$ref"].replace("#/components/schemas/", "")
            replacement = parent_schema[referent]
            # Substitute the reference before accessing the item type
            # prop_schema["items"] = replacement
            # I don't think it handles nested dataclass arrays!
            # Don't in fact even need to replace in the schema, just the var
            prop_schema = replacement
            # (We could dealias the entity at this point so as to title it?)
        prop_type = python_type(prop_schema["type"], prop_schema.get("format"))
        is_array_prop = prop_schema["type"] == "array"
        if is_substituted:
            prop_type = schema_mapping[referent]
        if is_array_prop and "items" in prop_schema:
            prop_array = prop_schema["items"]
            # TODO: I think arrays should be handled at level above, check this sooner?
            if "type" not in prop_array and prop_array and "$ref" in prop_array:
                referent = prop_array["$ref"].replace("#/components/schemas/", "")
                # TODO: Don't substitute: look up using Deserialisers Enum
                # replacement = parent_schema[referent]
                dto_enum_lookup = f"Deserialisers[{referent}].value"
                if schema_mapping is None:
                    item_type = dto_enum_lookup
                else:
                    item_type = schema_mapping.get(referent, dto_enum_lookup)
                    if is_pydantic:
                        # Important: stringify to postpone field's schema resolution
                        # (premature access will lead to PydanticSchemaGenerationError)
                        item_type = f'"{item_type}Model"'
                # Substitute the reference before accessing the item type
                # prop_schema["items"] = replacement
                # I don't think it handles nested dataclass arrays!
                # Don't in fact even need to replace in the schema, just the var
                # prop_array = replacement
                # (We could dealias the entity at this point so as to title it?)
            else:
                item_type = python_type(prop_array["type"], prop_array.get("format"))
            prop_type = f"{prop_type}[{item_type}]"
        elif is_pydantic and is_substituted:
            prop_type = f"{prop_type}Model"
        if prop_name in required:
            default = ""
        else:
            if is_array_prop:
                contains_list = True
                if is_pydantic:
                    default = ""
                else:
                    default = " = field(default_factory=list)"
            else:
                default = " = None"
        dc_source += f"    {to_pascal_case(prop_name)}: {prop_type}{default}\n"
    source_schema_default = hidden_field(schema_name, style=style)
    component_default = hidden_field(component_name, style=style)
    dc_source += f"    _source_schema_name: str = {source_schema_default}\n"
    dc_source += f"    _component_schema_name: str = {component_default}\n"
    jsonw_methods = """
    @classmethod
    def from_dict(cls, o):
        parent_schema = load_endpoint_component_schemas(cls._source_schema_name)
        schema = parent_schema[cls._component_schema_name]
        jsonschema.validate(o, schema)
        return fromdict(cls, o)

    class Meta(JSONWizard.Meta):
        key_transform_with_load = 'PASCAL'
        recursive_classes = True"""
    if is_jsonw:
        dc_source += jsonw_methods
    elif is_pydantic:
        # Patch field annotation bug https://github.com/pydantic/pydantic/issues/9093
        dc_source += f"\n\n{class_name}Model = {class_name}"
    jsonw_imports = {
        "dataclass_wizard": ["JSONWizard"],
        "dataclass_wizard.loaders": ["fromdict"],
        "jsonschema": [],
    }
    pydantic_imports = {
        "tubeulator.utils.string_conv": ["to_camel_case"],
        "pydantic": ["AliasGenerator", "BaseModel", "ConfigDict", "PrivateAttr"],
    }
    jsonw_utils = ["load_endpoint_component_schemas"]
    import_list = {
        "__future__": ["annotations"],
        "json": [],
        "datetime": ["datetime"],
        "dataclasses": ["dataclass"],
        # "pathlib": ["Path"], # I think this was a mistake
        **(jsonw_imports if is_jsonw else {}),
        **(pydantic_imports if is_pydantic else {}),
        "..utils.paths": [
            *(jsonw_utils if is_jsonw else []),
            "DtoEnum",
        ],
    }
    if contains_list or idx == 0:
        import_list["dataclasses"].append("field")
    imports = "\n".join(map(ast.unparse, starmap(import_node, import_list.items())))
    # Add imports here if generating multiple dataclasses [indicated by idx=0] (since we
    # can't see ahead), or if only making 1 [indicated by idx=None]
    return "\n\n".join([(imports if not idx else ""), dc_source])



from functools import lru_cache

from pymongo import MongoClient

from .mongod import MongodExceptionGuard


@lru_cache
def check_creds() -> dict[str, str]:
    """Store the API credentials in a simple singleton collection "tfl_cred" in the "creds"
    database if none has been stored there before. Interactive only.
    """
    with MongodExceptionGuard():
        client = MongoClient()
        creds_collection = client.creds.tfl_cred
        n_docs = creds_collection.estimated_document_count()
        if n_docs == 0:
            default_app_id = "tubeulator"
            app_id = (
                input(f"Enter your app ID (default: {default_app_id!r}): ")
                or default_app_id
            )
            primary_key = input("Enter your primary key: ")
            secondary_key = input("Enter your secondary key: ")
            stored_credential = {
                "app_id": app_id,
                "primary_key": primary_key,
                "secondary_key": secondary_key,
            }
            creds_collection.insert_one(stored_credential)
        elif n_docs > 1:
            raise ValueError(
                f"Multiple credentials {n_docs=} found: database has been corrupted",
            )
        else:
            stored_credential = creds_collection.find_one()
        return stored_credential



from .types import RouteEnum


class StopPointEndpointRoutes(RouteEnum):
    WITHIN_RADIUS = "/"
    """Gets a list of StopPoints within {radius} by the specified criteria"""
    FORWARD_REQUESTS = "/*"
    """Forwards any remaining requests to the back-end"""
    META_CATEGORIES = "/Meta/Categories"
    """Gets the list of available StopPoint additional information categories"""
    META_MODES = "/Meta/Modes"
    """Gets the list of available StopPoint modes"""
    META_STOP_TYPES = "/Meta/StopTypes"
    """Gets the list of available StopPoint types"""
    MODE = "/Mode/{modes}"
    """Gets a list of StopPoints filtered by the modes available at that StopPoint."""
    MODE_DISRUPTION = "/Mode/{modes}/Disruption"
    """Gets a distinct list of disrupted stop points for the given modes"""
    SEARCH = "/Search"
    """Search StopPoints by their common name, or their 5-digit Countdown Bus Stop Code."""
    SEARCH_QUERY = "/Search/{query}"
    """Search StopPoints by their common name, or their 5-digit Countdown Bus Stop Code."""
    SERVICE_TYPES = "/ServiceTypes"
    """Gets the service types for a given stoppoint"""
    SMS_ID = "/Sms/{id}"
    """Gets a StopPoint for a given sms code."""
    TYPE = "/Type/{types}"
    """Gets all stop points of a given type"""
    TYPE_PAGE = "/Type/{types}/page/{page}"
    """Gets all the stop points of given type(s) with a page number"""
    STOP_POINT_IDS = "/{ids}"
    """Gets a list of StopPoints corresponding to the given list of stop ids."""
    STOP_POINT_DISRUPTION = "/{ids}/Disruption"
    """Gets all disruptions for the specified StopPointId, plus disruptions for any
    child Naptan records it may have."""
    ARRIVAL_DEPARTURES = "/{id}/ArrivalDepartures"
    """Gets the list of arrival and departure predictions for the given stop point id
    (overground and tfl rail only)"""
    ARRIVALS = "/{id}/Arrivals"
    """Gets the list of arrival predictions for the given stop point id"""
    CAN_REACH_ON_LINE = "/{id}/CanReachOnLine/{lineId}"
    """Gets Stopoints that are reachable from a station/line combination."""
    CROWDING = "/{id}/Crowding/{line}"
    """Gets all the Crowding data (static) for the StopPointId, plus crowding data
    for a given line and optionally a particular direction."""
    DIRECTION_TO = "/{id}/DirectionTo/{toStopPointId}"
    """Returns the canonical direction, "inbound" or "outbound", for a given pair of
    stop point Ids in the direction from -&gt; to."""
    ROUTE = "/{id}/Route"
    """Returns the route sections for all the lines that service the given stop point
    ids"""
    PLACE_TYPES = "/{id}/placeTypes"
    """Get a list of places corresponding to a given id and place types."""
    CAR_PARKS = "/{stopPointId}/CarParks"
    """Get car parks corresponding to the given stop point id."""
    TAXI_RANKS = "/{stopPointId}/TaxiRanks"
    """Gets a list of taxi ranks corresponding to the given stop point id."""



"""Import this module to check if the generated Python DTOs are loadable."""

try:
    from ..generated import (
        AccidentStats,
        AirQuality,
        BikePoint,
        DisruptionsLiftsv2,
        Journey,
        Line,
        Mode,
        Place,
        Road,
        Search,
        StopPoint,
        Vehicle,
        crowding,
        occupancy,
    )
except:
    # The load_test is used for parsing, but can block generation
    # Temporarily use pass here to permit regenerating from broken config
    raise  # pass



from __future__ import annotations

import logging
from copy import deepcopy
from pprint import pprint

from .._types import ApiAliasToUnifiedEntities, ApiEntityAlias
from ..reference import AliasToRefs, Reference, dealias_schema
from .base import ApiSchema
from .unified import single_unified_api_schema

__all__ = ["AliasedApiSchema"]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class AliasedApiSchema(ApiSchema):
    """A non-unified API schema."""

    unified: bool = False

    def __post_init__(self):
        super().__post_init__()
        self.prepare_inventories()
        self.match_entities()
        # self.pprint_api_inventory(nonreferential_ok=True)
        self.chase_entity_references()
        if not self.is_resolved:
            self.property_match()

    @property
    def is_resolved(self) -> bool:
        """If there is no alias in the schema with empty entity list (i.e. unresolved)."""
        return all(self.alias2ents.values())
        # return all(
        #     {
        #         name: entities
        #         for name, entities in self.alias2ents.items()
        #         if not self.is_skipped_name(name=name)
        #     }.values(),
        # )

    def match_entities(self) -> None:
        """First pass, get the exact matches. Iterate over the aliases (the entity names in
        the API schema "broken out" from the unified schema) and where the schema
        component is an exact match for one in the unified schema, store the entity name
        (or names: there is no guarantee only a single entity will match the alias).
        """
        for entity_alias, schema_component in self.entity_schemas.items():
            resolved_ents = []
            for entity, unif_schema in self.unified_schema.entity_schemas.items():
                if schema_component == unif_schema:
                    resolved_ents.append(entity)
            if self.alias2ents[entity_alias] == [] and entity_alias in resolved_ents:
                # The entities with Array in their name will match any other Array,
                # but it's trivially always just going to be the one with same name
                resolved_ents = [entity_alias]
            for entity in resolved_ents:
                self.resolve_alias(alias=entity_alias, entity=entity)

    def resolve_alias(self, alias: ApiEntityAlias, entity: str) -> None:
        if entity not in (entity_list := self.alias2ents[alias]):
            entity_list.append(entity)

    def match_entities_with_modification(
        self,
        alias: ApiEntityAlias,
        property_refs: list[Reference],
    ) -> None:
        """Second pass, get the exact matches after substituting a reference.  Iterate over
        the aliases (the entity names in the API schema "broken out" from the unified
        schema) and where the schema component is an exact match for one in the unified
        schema, store the entity name (or names: there is no guarantee only a single
        entity will match the alias).
        """
        schema_component = self.entity_schemas[alias]
        # Modify the schema dict by replacing the reference (in a copy not the original)
        mod_schema_component = dealias_schema(
            refs=property_refs,
            component=schema_component,
        )
        for entity, unif_schema in self.unified_schema.entity_schemas.items():
            if mod_schema_component == unif_schema:
                self.resolve_alias(alias=alias, entity=entity)

    def chase_entity_references(self) -> None:
        """Run the entity dealiasing/cross-reference in a loop until no more are resolved."""
        iteration_count = 0
        while any((resolution_counts := self.resolve_entity_references()).values()):
            iteration_count += 1
            if iteration_count > 10:
                raise ValueError("Looped too much, something's up")
            for alias, refs in self.resolved_property_refs.items():
                if refs:
                    self.match_entities_with_modification(
                        alias=alias,
                        property_refs=refs,
                    )
            # Keep the running tally of resolved references for each alias up to date
            for alias in self.resolution_counts:
                self.resolution_counts[alias] += resolution_counts[alias]
        else:
            logger.info(
                f"-----RESOLUTION COMPLETE FOR {self.name}: "
                f" ROUNDS: {iteration_count}, "
                f" REFERENTIAL: {self.any_referential_properties}----------",
            )

    def resolve_entity_references(self) -> dict[ApiEntityAlias, int]:
        """Second pass, resolve the referential matches with substitution of known values.
        Again iterate over the aliases, but now try to find where the schema component
        can be substituted if the reference is 'completable' to consequently make an
        exact match for one of the schema components in the unified schema. As for the
        previous pass, store the entity name or names alongside the alias if successful.
        """
        resolution_counts = dict.fromkeys(self.entity_schemas, 0)
        for entity_alias, schema_component in self.entity_schemas.items():
            if properties := schema_component.get("properties"):
                # assert self.property_refs[entity_alias] == [], "Expected fresh list"
                # IMPORTANT: listcomp is used here so list is fixed at this point
                skip_refs = list(self.property_refs[entity_alias])
                """Skip any properties whose references have been resolved already"""
                # Either a top-level $ref or as the items in an array.
                # (Valid) arrays have items with a specified $ref
                property_refs = [
                    # Use next as can rely on only 1 reference per property
                    next(
                        Reference(
                            schema=self.name,
                            alias=entity_alias,
                            referential_property=prop,
                            is_array=(is_nested_ref := isinstance(v, dict)),
                            entity=v["$ref"] if is_nested_ref else v,
                        )
                        for k, v in prop_type_info.items()
                        if k in ["$ref", "items"]
                    )
                    for prop, prop_type_info in properties.items()
                    if (
                        "$ref" in prop_type_info
                        or "$ref" in prop_type_info.get("items", {})
                    )
                    if prop not in [ref.referential_property for ref in skip_refs]
                ]
                logger.debug(f"[{entity_alias}] Adding {property_refs}")
                self.property_refs[entity_alias].extend(property_refs)
                # self.pprint_property_refs(alias=entity_alias)
                n_resolved = self.resolve_refs(alias=entity_alias, skip=skip_refs)
                resolution_counts[entity_alias] += n_resolved
        return resolution_counts

    def pprint_property_refs(self, alias: str) -> None:
        if property_refs := self.property_refs[alias]:
            pprint(property_refs)
            completability = [
                property_ref.is_completable(api_inventory=self.alias2ents)
                for property_ref in property_refs
            ]
            print(
                f"Completable: {completability.count(True)}, incomplete: {completability.count(False)}",
            )

    def resolve_refs(self, alias: str, skip: list[Reference]) -> int:
        """Resolve any 'completable' references, and return the number that were resolved,
        making it possible to ``while`` loop over this method to resolve all references.
        """
        resolved_property_refs = []
        if property_refs := self.property_refs[alias]:
            for property_ref in property_refs:
                if property_ref in skip:
                    continue
                elif property_ref.is_completable(api_inventory=self.alias2ents):
                    completed = self.dealias_property_reference(ref=property_ref)
                    resolved_property_refs.append(completed)
            logger.debug(f"Resolved property refs: {len(resolved_property_refs)}")
            self.resolved_property_refs[alias].extend(resolved_property_refs)
        return len(resolved_property_refs)

    def pprint_api_inventory(self, nonreferential_ok: bool = False) -> None:
        """Print the API inventory and a newline if any referential properties printed."""
        if self.any_referential_properties or nonreferential_ok:
            pprint({k: v for k, v in self.alias2ents.items() if "Response" not in k})
            print()

    def dealias_property_reference(self, ref: Reference) -> Reference:
        """Substitute the aliased entity, which has the format
        ``"#/components/schemas/{ref.alias}"``, with the proper entity name from the
        unified API, which must be 'completable' i.e. exist in the ``self.alias2ents``
        dictionary.

        The new :obj:`Reference` will then suffice to modify the API schema property
        dictionary so as to make it identical to one in the unified API, and thereby
        facilitate matching, and thus resolving the alias to its proper name.
        """
        # Take the first item in the list: no way to distinguish b/w multiple options
        # (if this doesn't work then consider iterating over possible multiple options)
        dealiased_ref_entity = self.alias2ents[ref.substituted_ref][0]
        new_entity = ref.entity[: -len(ref.substituted_ref)] + dealiased_ref_entity
        dealiased_ref = Reference(
            schema=ref.schema,
            alias=ref.alias,
            referential_property=ref.referential_property,
            is_array=ref.is_array,
            entity=new_entity,
        )
        return dealiased_ref

    def property_match(self) -> None:
        """After the reference-based alias-to-entity resolution, resolve remaining aliases
        by matching the properties of each alias to those of unified API definitions.

        Can match via total number of properties/property names/property descriptions.
        """
        for entity_alias in self.entity_schemas:
            if self.alias2ents[entity_alias] != []:
                continue
            property_count = next(
                count
                for count, entities in self.ents_by_property_count.items()
                if entity_alias in entities
            )
            candidates = self.unified_schema.ents_by_property_count[property_count]
            if len(candidates) > 1:
                if property_count > 0:
                    properties_to_match = list(
                        self.entity_schemas[entity_alias]["properties"],
                    )
                    property_lookup = {
                        entity: self.unified_schema.entity_schemas[entity]["properties"]
                        for entity in candidates
                    }
                    matched = {
                        entity: properties
                        for entity, properties in property_lookup.items()
                        if list(properties) == properties_to_match
                    }
                    if len(matched) > 0:
                        candidates = matched
            # else IDK
            if len(candidates) == 1:
                entity_name = next(iter(candidates))
                logger.debug(f"[PM] Established {entity_alias} == {entity_name}")
                self.resolve_alias(alias=entity_alias, entity=entity_name)

    def prepare_inventories(self) -> None:
        """Set the ``unified_schema`` as the unified API schema.

        Set up empty inventories keyed by the source entity alias:
        - to lists of the 'proper' entity names in the unified schema
        - to lists of the schema property references (initial)
        - to lists of the schema property references ('resolved' or 'dealiased')
        """
        self.unified_schema = single_unified_api_schema
        self.alias2ents: ApiAliasToUnifiedEntities = {
            alias: [] for alias in self.entity_schemas
        }
        self.property_refs: AliasToRefs = {alias: [] for alias in self.entity_schemas}
        self.resolved_property_refs: AliasToRefs = deepcopy(self.property_refs)
        self.resolution_counts = dict.fromkeys(self.entity_schemas, 0)
        """
        Extends the property refs lists for each alias in ``property_refs`` dict,
        then when ``resolve_refs`` is called for the alias which dealiases
        the property reference for any that are completable, stores the
        ``resolved_property_refs`` in a separate dict similar to
        ``property_refs``, then returns the number of refs 'completed'/resolved.
        """

    @property
    def any_referential_properties(self) -> bool:
        """If any of the API schema component properties included a reference to another
        entity, then the values of the ``property_refs`` dict will be non-empty.
        """
        return any(v for v in self.property_refs.values())



from .types import RouteEnum


class PlaceEndpointRoutes(RouteEnum):
    PLACES_BY_GEO_REGION = "/"
    """Gets the places that lie within a geographic region. The geographic region of interest can either
    be specified by using a lat/lon geo-point and a radius in metres to return places within the locus
    defined by the lat/lon of its centre or alternatively, by the use of a bounding box defined by the
    lat/lon of its north-west and south-east corners. Optionally filters on type and can strip
    properties for a smaller payload."""
    FORWARD_REQUESTS = "/*"
    """Forwards any remaining requests to the back-end"""
    META_CATEGORIES = "/Meta/Categories"
    """Gets a list of all of the available place property categories and keys."""
    META_PLACE_TYPES = "/Meta/PlaceTypes"
    """Gets a list of the available types of Place."""
    SEARCH_PLACES = "/Search"
    """Gets all places that matches the given query"""
    PLACES_BY_TYPE = "/Type/{types}"
    """Gets all places of a given type"""
    PLACE_BY_ID = "/{id}"
    """Gets the place with the given id."""
    PLACE_BY_TYPE_AT_COORDINATES = "/{type}/At/{lat}/{lon}"
    """Gets any places of the given type whose geography intersects the given latitude and longitude. In
    practice this means the Place must be polygonal e.g. a BoroughBoundary."""



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class Mode(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Mode
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    IsTflService: bool = None
    IsFarePaying: bool = None
    IsScheduledService: bool = None
    ModeName: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl')


ModeModel = Mode


class PathAttribute(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.PathAttribute
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Name: str = None
    Value: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-2')


PathAttributeModel = PathAttribute


class InstructionStep(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.InstructionStep
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Description: str = None
    TurnDirection: str = None
    StreetName: str = None
    Distance: int = None
    CumulativeDistance: int = None
    SkyDirection: int = None
    SkyDirectionDescription: str = None
    CumulativeTravelTime: int = None
    Latitude: float = None
    Longitude: float = None
    PathAttribute: PathAttributeModel = None
    DescriptionHeading: str = None
    TrackType: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-3')


InstructionStepModel = InstructionStep


class Instruction(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Instruction
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Summary: str = None
    Detailed: str = None
    Steps: list["InstructionStepModel"]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-4')


InstructionModel = Instruction


class Obstacle(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.Obstacle
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Type: str = None
    Incline: str = None
    StopId: int = None
    Position: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-5')


ObstacleModel = Obstacle


class Point(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Point
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Lat: float = None
    Lon: float = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-6')


PointModel = Point


class PassengerFlow(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.PassengerFlow
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    TimeSlice: str = None
    Value: int = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-7')


PassengerFlowModel = PassengerFlow


class TrainLoading(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.TrainLoading
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Line: str = None
    LineDirection: str = None
    PlatformDirection: str = None
    Direction: str = None
    NaptanTo: str = None
    TimeSlice: str = None
    Value: int = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-8')


TrainLoadingModel = TrainLoading


class Crowding(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Crowding
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    PassengerFlows: list["PassengerFlowModel"]
    TrainLoadings: list["TrainLoadingModel"]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-9')


CrowdingModel = Crowding


class Identifier(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Identifier
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Name: str = None
    Uri: str = None
    FullName: str = None
    Type: str = None
    Crowding: CrowdingModel = None
    RouteType: str = None
    Status: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-10')


IdentifierModel = Identifier


class JpElevation(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Common.JourneyPlanner.JpElevation
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Distance: int = None
    StartLat: float = None
    StartLon: float = None
    EndLat: float = None
    EndLon: float = None
    HeightFromPreviousPoint: int = None
    Gradient: float = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-11')


JpElevationModel = JpElevation


class Path(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.Path
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    LineString: str = None
    StopPoints: list["IdentifierModel"]
    Elevation: list["JpElevationModel"]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-12')


PathModel = Path


class RouteOption(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.RouteOption
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Name: str = None
    Directions: list[str]
    LineIdentifier: IdentifierModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-13')


RouteOptionModel = RouteOption


class LineGroup(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.LineGroup
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    NaptanIdReference: str = None
    StationAtcoCode: str = None
    LineIdentifier: list[str]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-14')


LineGroupModel = LineGroup


class LineModeGroup(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.LineModeGroup
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    ModeName: str = None
    LineIdentifier: list[str]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-15')


LineModeGroupModel = LineModeGroup


class AdditionalProperties(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.AdditionalProperties
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Category: str = None
    Key: str = None
    SourceSystemKey: str = None
    Value: str = None
    Modified: datetime = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-16')


AdditionalPropertiesModel = AdditionalProperties


class Place(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Place
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Url: str = None
    CommonName: str = None
    Distance: float = None
    PlaceType: str = None
    AdditionalProperties: list["AdditionalPropertiesModel"]
    Children: list["PlaceModel"]
    ChildrenUrls: list[str]
    Lat: float = None
    Lon: float = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-17')


PlaceModel = Place


class StopPoint(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.StopPoint
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    NaptanId: str = None
    PlatformName: str = None
    Indicator: str = None
    StopLetter: str = None
    Modes: list[str]
    IcsCode: str = None
    SmsCode: str = None
    StopType: str = None
    StationNaptan: str = None
    AccessibilitySummary: str = None
    HubNaptanCode: str = None
    Lines: list["IdentifierModel"]
    LineGroup: list["LineGroupModel"]
    LineModeGroups: list["LineModeGroupModel"]
    FullName: str = None
    NaptanMode: str = None
    Status: bool = None
    Id: str = None
    Url: str = None
    CommonName: str = None
    Distance: float = None
    PlaceType: str = None
    AdditionalProperties: list["AdditionalPropertiesModel"]
    Children: list["PlaceModel"]
    ChildrenUrls: list[str]
    Lat: float = None
    Lon: float = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-18')


StopPointModel = StopPoint


class RouteSectionNaptanEntrySequence(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.RouteSectionNaptanEntrySequence
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Ordinal: int = None
    StopPoint: StopPointModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-19')


RouteSectionNaptanEntrySequenceModel = RouteSectionNaptanEntrySequence


class RouteSection(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.RouteSection
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    LineId: str = None
    RouteCode: str = None
    Name: str = None
    LineString: str = None
    Direction: str = None
    OriginationName: str = None
    DestinationName: str = None
    ValidTo: datetime = None
    ValidFrom: datetime = None
    RouteSectionNaptanEntrySequence: list["RouteSectionNaptanEntrySequenceModel"]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-20')


RouteSectionModel = RouteSection


class Disruption(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Disruption
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Category: str = None
    Type: str = None
    CategoryDescription: str = None
    Description: str = None
    Summary: str = None
    AdditionalInfo: str = None
    Created: datetime = None
    LastUpdate: datetime = None
    AffectedRoutes: list["RouteSectionModel"]
    AffectedStops: list["StopPointModel"]
    ClosureText: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-21')


DisruptionModel = Disruption


class PlannedWork(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.PlannedWork
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Description: str = None
    CreatedDateTime: datetime = None
    LastUpdateDateTime: datetime = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-22')


PlannedWorkModel = PlannedWork


class Leg(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.Leg
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Duration: int = None
    Speed: str = None
    Instruction: InstructionModel = None
    Obstacles: list["ObstacleModel"]
    DepartureTime: datetime = None
    ArrivalTime: datetime = None
    DeparturePoint: PointModel = None
    ArrivalPoint: PointModel = None
    Path: PathModel = None
    RouteOptions: list["RouteOptionModel"]
    Mode: IdentifierModel = None
    Disruptions: list["DisruptionModel"]
    PlannedWorks: list["PlannedWorkModel"]
    Distance: float = None
    IsDisrupted: bool = None
    HasFixedLocations: bool = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-23')


LegModel = Leg


class FareTapDetails(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.FareTapDetails
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    ModeType: str = None
    ValidationType: str = None
    HostDeviceType: str = None
    BusRouteId: str = None
    NationalLocationCode: int = None
    TapTimestamp: datetime = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-24')


FareTapDetailsModel = FareTapDetails


class FareTap(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.FareTap
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    AtcoCode: str = None
    TapDetails: FareTapDetailsModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-25')


FareTapModel = FareTap


class Fare(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.Fare
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    LowZone: int = None
    HighZone: int = None
    Cost: int = None
    ChargeProfileName: str = None
    IsHopperFare: bool = None
    ChargeLevel: str = None
    Peak: int = None
    OffPeak: int = None
    Taps: list["FareTapModel"]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-26')


FareModel = Fare


class FareCaveat(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.FareCaveat
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Text: str = None
    Type: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-27')


FareCaveatModel = FareCaveat


class JourneyFare(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.JourneyFare
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    TotalCost: int = None
    Fares: list["FareModel"]
    Caveats: list["FareCaveatModel"]
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-28')


JourneyFareModel = JourneyFare


class Journey(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.Journey
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    StartDateTime: datetime = None
    Duration: int = None
    ArrivalDateTime: datetime = None
    Legs: list["LegModel"]
    Fare: JourneyFareModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-29')


JourneyModel = Journey


class ValidityPeriod(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.ValidityPeriod
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    FromDate: datetime = None
    ToDate: datetime = None
    IsNow: bool = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-30')


ValidityPeriodModel = ValidityPeriod


class LineStatus(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.LineStatus
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: int = None
    LineId: str = None
    StatusSeverity: int = None
    StatusSeverityDescription: str = None
    Reason: str = None
    Created: datetime = None
    Modified: datetime = None
    ValidityPeriods: list["ValidityPeriodModel"]
    Disruption: DisruptionModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-31')


LineStatusModel = LineStatus


class MatchedRoute(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.MatchedRoute
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    RouteCode: str = None
    Name: str = None
    Direction: str = None
    OriginationName: str = None
    DestinationName: str = None
    Originator: str = None
    Destination: str = None
    ServiceType: str = None
    ValidTo: datetime = None
    ValidFrom: datetime = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-32')


MatchedRouteModel = MatchedRoute


class LineServiceTypeInfo(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.LineServiceTypeInfo
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Name: str = None
    Uri: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-33')


LineServiceTypeInfoModel = LineServiceTypeInfo


class Line(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.Line
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: str = None
    Name: str = None
    ModeName: str = None
    Disruptions: list["DisruptionModel"]
    Created: datetime = None
    Modified: datetime = None
    LineStatuses: list["LineStatusModel"]
    RouteSections: list["MatchedRouteModel"]
    ServiceTypes: list["LineServiceTypeInfoModel"]
    Crowding: CrowdingModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-34')


LineModel = Line


class JourneyPlannerCycleHireDockingStationData(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.JourneyPlannerCycleHireDockingStationData
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    OriginNumberOfBikes: int = None
    DestinationNumberOfBikes: int = None
    OriginNumberOfEmptySlots: int = None
    DestinationNumberOfEmptySlots: int = None
    OriginId: str = None
    DestinationId: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-35')


JourneyPlannerCycleHireDockingStationDataModel = JourneyPlannerCycleHireDockingStationData


class TimeAdjustment(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.TimeAdjustment
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Date: str = None
    Time: str = None
    TimeIs: str = None
    Uri: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-36')


TimeAdjustmentModel = TimeAdjustment


class TimeAdjustments(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.TimeAdjustments
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Earliest: TimeAdjustmentModel = None
    Earlier: TimeAdjustmentModel = None
    Later: TimeAdjustmentModel = None
    Latest: TimeAdjustmentModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-37')


TimeAdjustmentsModel = TimeAdjustments


class SearchCriteria(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.SearchCriteria
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    DateTime: datetime = None
    DateTimeType: str = None
    TimeAdjustments: TimeAdjustmentsModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-38')


SearchCriteriaModel = SearchCriteria


class JourneyVector(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.JourneyVector
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    From: str = None
    To: str = None
    Via: str = None
    Uri: str = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-39')


JourneyVectorModel = JourneyVector


class ItineraryResult(BaseModel):
    """
    Autogenerated from Journey::Tfl.Api.Presentation.Entities.JourneyPlanner.ItineraryResult
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Journeys: list["JourneyModel"]
    Lines: list["LineModel"]
    CycleHireDockingStationData: JourneyPlannerCycleHireDockingStationDataModel = None
    StopMessages: list[str]
    RecommendedMaxAgeMinutes: int = None
    SearchCriteria: SearchCriteriaModel = None
    JourneyVector: JourneyVectorModel = None
    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Tfl-40')


ItineraryResultModel = ItineraryResult


class MetaModesGet200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Journey::MetaModesGet200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='MetaModesGet200ApplicationJsonResponse')


MetaModesGet200ApplicationJsonResponseModel = MetaModesGet200ApplicationJsonResponse


class Get200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Journey::Get200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Journey')
    _component_schema_name: str = PrivateAttr(default='Get200ApplicationJsonResponse')


Get200ApplicationJsonResponseModel = Get200ApplicationJsonResponse


class Deserialisers(DtoEnum):
    Tfl = Mode
    Tfl_2 = PathAttribute
    Tfl_3 = InstructionStep
    Tfl_4 = Instruction
    Tfl_5 = Obstacle
    Tfl_6 = Point
    Tfl_7 = PassengerFlow
    Tfl_8 = TrainLoading
    Tfl_9 = Crowding
    Tfl_10 = Identifier
    Tfl_11 = JpElevation
    Tfl_12 = Path
    Tfl_13 = RouteOption
    Tfl_14 = LineGroup
    Tfl_15 = LineModeGroup
    Tfl_16 = AdditionalProperties
    Tfl_17 = Place
    Tfl_18 = StopPoint
    Tfl_19 = RouteSectionNaptanEntrySequence
    Tfl_20 = RouteSection
    Tfl_21 = Disruption
    Tfl_22 = PlannedWork
    Tfl_23 = Leg
    Tfl_24 = FareTapDetails
    Tfl_25 = FareTap
    Tfl_26 = Fare
    Tfl_27 = FareCaveat
    Tfl_28 = JourneyFare
    Tfl_29 = Journey
    Tfl_30 = ValidityPeriod
    Tfl_31 = LineStatus
    Tfl_32 = MatchedRoute
    Tfl_33 = LineServiceTypeInfo
    Tfl_34 = Line
    Tfl_35 = JourneyPlannerCycleHireDockingStationData
    Tfl_36 = TimeAdjustment
    Tfl_37 = TimeAdjustments
    Tfl_38 = SearchCriteria
    Tfl_39 = JourneyVector
    Tfl_40 = ItineraryResult
    MetaModesGet200ApplicationJsonResponse = MetaModesGet200ApplicationJsonResponse
    Get200ApplicationJsonResponse = Get200ApplicationJsonResponse


from .types import RouteEnum


class JourneyEndpointRoutes(RouteEnum):
    FORWARD_REQUESTS = "/*"
    """Forwards any remaining requests to the back-end"""
    JOURNEY_RESULTS = "/JourneyResults/{from}/to/{to}"
    """Perform a Journey Planner search from the parameters specified in simple types"""
    MODES = "/Meta/Modes"
    """Gets a list of all of the available journey planner modes"""



from .types import RouteEnum


class CrowdingEndpointRoutes(RouteEnum):
    NAPTAN_CROWDING = "/{Naptan}"
    """Returns crowding information for Naptan"""
    LIVE_CROWDING = "/{Naptan}/Live"
    """Returns live crowding information for Naptan"""
    DOW_CROWDING = "/{Naptan}/{DayOfWeek}"
    """Returns crowding information for Naptan for Day of Week"""



from .types import RouteEnum


class LineEndpointRoutes(RouteEnum):
    FORWARD_REQUESTS = "/*"
    """Forwards any remaining requests to the back-end"""
    META_DISRUPTION_CATEGORIES = "/Meta/DisruptionCategories"
    """Gets a list of valid disruption categories"""
    META_MODES = "/Meta/Modes"
    """Gets a list of valid modes"""
    META_SERVICE_TYPES = "/Meta/ServiceTypes"
    """Gets a list of valid ServiceTypes to filter on"""
    META_SEVERITY = "/Meta/Severity"
    """Gets a list of valid severity codes"""
    LINES_BY_MODES = "/Mode/{modes}"
    """Gets lines that serve the given modes."""
    DISRUPTION_BY_MODES = "/Mode/{modes}/Disruption"
    """Get disruptions for all lines of the given modes."""
    ROUTE_BY_MODES = "/Mode/{modes}/Route"
    """Gets all lines and their valid routes for given modes, including the name and id of
    the originating and terminating stops for each route"""
    STATUS_BY_MODES = "/Mode/{modes}/Status"
    """Gets the line status of for all lines for the given modes"""
    ALL_ROUTES = "/Route"
    """Get all valid routes for all lines, including the name and id of the originating and
    terminating stops for each route."""
    SEARCH_LINES_ROUTES = "/Search/{query}"
    """Search for lines or routes matching the query string"""
    STATUS_BY_SEVERITY = "/Status/{severity}"
    """Gets the line status for all lines with a given severity. A list of valid severity
    codes can be obtained from a call to Line/Meta/Severity"""
    LINES_BY_IDS = "/{ids}"
    """Gets lines that match the specified line ids."""
    ARRIVALS_BY_IDS = "/{ids}/Arrivals"
    """Get the list of arrival predictions for given line ids based at the given stop"""
    ARRIVALS_BY_IDS_STOP = "/{ids}/Arrivals/{stopPointId}"
    """Get the list of arrival predictions for given line ids based at the given stop"""
    DISRUPTION_BY_IDS = "/{ids}/Disruption"
    """Get disruptions for the given line ids"""
    ROUTE_BY_IDS = "/{ids}/Route"
    """Get all valid routes for given line ids, including the name and id of the
    originating and terminating stops for each route."""
    STATUS_BY_IDS = "/{ids}/Status"
    """Gets the line status of for given line ids e.g Minor Delays"""
    STATUS_BY_IDS_DATES = "/{ids}/Status/{startDate}/to/{endDate}"
    """Gets the line status for given line ids during the provided dates e.g Minor Delays"""
    ROUTE_SEQUENCE_BY_ID_DIRECTION = "/{id}/Route/Sequence/{direction}"
    """Gets all valid routes for given line id, including the sequence of stops on each
    route."""
    STOP_POINTS_BY_ID = "/{id}/StopPoints"
    """Gets a list of the stations that serve the given line id"""
    TIMETABLE_BY_ID_FROM_STOP = "/{id}/Timetable/{fromStopPointId}"
    """Gets the timetable for a specified station on the give line"""
    TIMETABLE_BY_ID_FROM_TO_STOP = (
        "/{id}/Timetable/{fromStopPointId}/to/{toStopPointId}"
    )
    """Gets the timetable for a specified station on the give line with specified
    destination"""



from __future__ import annotations

from dataclasses import dataclass, field

from .utils.logs import set_up_logging

__all__ = ["Journey", "Place", "Destination", "Departure"]

logger = set_up_logging(name=__name__)


@dataclass
class Journey:
    A: str
    B: str
    via: list[str] = field(default_factory=list)


class Place:
    ...


@dataclass
class Destination(Place):
    ...


@dataclass
class Departure(Place):
    ...



from __future__ import annotations

from ...utils.paths import unified_api_schema as unified_api_schema_path
from .base import ApiSchema

__all__ = [
    "UnifiedApiSchema",
    "single_unified_api_schema",
]


class UnifiedApiSchema(ApiSchema):
    """A unified API schema."""

    unified: bool = True


single_unified_api_schema = UnifiedApiSchema(path=unified_api_schema_path, unified=True)
"""
A singleton instance to avoid re-reading the unified API schema JSON once per
non-unified API schema instantiation.
"""



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class AccidentStatsAccidentDetailArray(BaseModel):
    """
    Autogenerated from AccidentStats::Tfl-Api-Presentation-Entities-AccidentStats-AccidentDetailArray
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='AccidentStats')
    _component_schema_name: str = PrivateAttr(default='Tfl-Api-Presentation-Entities-AccidentStats-AccidentDetailArray')


AccidentStatsAccidentDetailArrayModel = AccidentStatsAccidentDetailArray


class AccidentDetail(BaseModel):
    """
    Autogenerated from AccidentStats::Tfl.Api.Presentation.Entities.AccidentStats.AccidentDetail
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Id: int = None
    Lat: float = None
    Lon: float = None
    Location: str = None
    Date: datetime = None
    Severity: str = None
    Borough: str = None
    Casualties: list["CasualtyModel"]
    Vehicles: list["VehicleModel"]
    _source_schema_name: str = PrivateAttr(default='AccidentStats')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.AccidentStats.AccidentDetail')


AccidentDetailModel = AccidentDetail


class Casualty(BaseModel):
    """
    Autogenerated from AccidentStats::Tfl.Api.Presentation.Entities.AccidentStats.Casualty
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Age: int = None
    Class: str = None
    Severity: str = None
    Mode: str = None
    AgeBand: str = None
    _source_schema_name: str = PrivateAttr(default='AccidentStats')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.AccidentStats.Casualty')


CasualtyModel = Casualty


class Vehicle(BaseModel):
    """
    Autogenerated from AccidentStats::Tfl.Api.Presentation.Entities.AccidentStats.Vehicle
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    Type: str = None
    _source_schema_name: str = PrivateAttr(default='AccidentStats')
    _component_schema_name: str = PrivateAttr(default='Tfl.Api.Presentation.Entities.AccidentStats.Vehicle')


VehicleModel = Vehicle


class Deserialisers(DtoEnum):
    Tfl_Api_Presentation_Entities_AccidentStats_AccidentDetailArray = AccidentStatsAccidentDetailArray
    Tfl__Api__Presentation__Entities__AccidentStats__AccidentDetail = AccidentDetail
    Tfl__Api__Presentation__Entities__AccidentStats__Casualty = Casualty
    Tfl__Api__Presentation__Entities__AccidentStats__Vehicle = Vehicle


from .types import RouteEnum


class ModeEndpointRoutes(RouteEnum):
    ACTIVE_SERVICE_TYPES = "/ActiveServiceTypes"
    """Returns the service type active for a mode. Currently only supports tube"""
    MODE_ARRIVALS = "/{mode}/Arrivals"
    """Gets the next arrival predictions for all stops of a given mode"""



from __future__ import annotations
import json
from datetime import datetime
from dataclasses import dataclass, field
from tubeulator.utils.string_conv import to_camel_case
from pydantic import AliasGenerator, BaseModel, ConfigDict, PrivateAttr
from ..utils.paths import DtoEnum

class LiftDisruption(BaseModel):
    """
    Autogenerated from Disruptions-Lifts-v2::LiftDisruption
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    IcsCode: str = None
    NaptanCode: str = None
    StopPointName: str = None
    OutageStartArea: str = None
    OutageEndArea: str = None
    Message: str = None
    _source_schema_name: str = PrivateAttr(default='Disruptions-Lifts-v2')
    _component_schema_name: str = PrivateAttr(default='LiftDisruption')


LiftDisruptionModel = LiftDisruption


class Get200ApplicationJsonResponse(BaseModel):
    """
    Autogenerated from Disruptions-Lifts-v2::Get200ApplicationJsonResponse
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(validation_alias=to_camel_case),
    )

    _source_schema_name: str = PrivateAttr(default='Disruptions-Lifts-v2')
    _component_schema_name: str = PrivateAttr(default='Get200ApplicationJsonResponse')


Get200ApplicationJsonResponseModel = Get200ApplicationJsonResponse


class Deserialisers(DtoEnum):
    LiftDisruption = LiftDisruption
    Get200ApplicationJsonResponse = Get200ApplicationJsonResponse


import json
from dataclasses import dataclass, field
from enum import Enum
from functools import cache
from pathlib import Path

from .. import __path__

__all__ = [
    "pkg_path",
    "db_path",
    "openapi_schemas_path",
    "openapi_unified_path",
    "find_schema",
    "find_schema_by_name",
    "load_endpoint_schema",
    "load_endpoint_component_schemas",
    "RefPath",
    "SchemaPath",
    "DtoEnum",
    "unified_api_schema",
]

pkg_path = Path(*__path__)
data_path = pkg_path / "data"
db_path = data_path / "db"
openapi_schemas_path = data_path / "openapi"
openapi_unified_path = data_path / "openapi_unified"


def find_schema(schema_dir: Path, match: str = "*", levels: int = 0) -> Path:
    """Get the next schema after traversing the given number of levels (default: no levels,
    i.e. directly in the given directory). The schema must have a ``.json`` suffix.
    """
    try:
        glob_prefix = "*/" * levels
        glob_pattern = f"{glob_prefix}{match}.json"
        return next(schema_dir.glob(glob_pattern))
    except StopIteration:
        raise ValueError(f"No JSON file {levels=} below {schema_dir}")


@cache
def find_schema_by_name(schema_name: str) -> Path:
    return find_schema(schema_dir=openapi_schemas_path, match=schema_name, levels=1)


@cache
def load_endpoint_schema(schema_name: str):
    """Load an entire JSON schema for an API endpoint by its name, e.g. "Line" or "Mode"."""
    endpoint_schema = json.loads(Path(find_schema_by_name(schema_name)).read_text())
    return endpoint_schema


@cache
def load_endpoint_component_schemas(schema_name: str) -> dict[str, dict]:
    """Load all component schemas of a JSON schema for an API endpoint by endpoint name."""
    endpoint_schema = load_endpoint_schema(schema_name)
    component_schemas = endpoint_schema["components"].get("schemas", {})
    return component_schemas


@dataclass
class RefPath:
    path: str = field(repr=False)
    name: str = field(init=False)

    def __post_init__(self):
        self.name = Path(self.path).name


@dataclass
class SchemaPath:
    source: dict[str, str] = field(repr=False)
    ref: RefPath = field(init=False)

    def __post_init__(self):
        ref = self.source.get("items", {}).get("$ref")
        self.ref = RefPath(ref)


def to_enum_friendly_str(name: str) -> str:
    return name.replace("-", "_").replace(".", "__")


def from_enum_friendly_str(name: str) -> str:
    return name.replace("_", "-").replace("__", ".")


class DtoEnum(Enum):
    @classmethod
    def select_component(cls, component_name: str):
        return cls[to_enum_friendly_str(component_name)]


unified_api_schema = find_schema(schema_dir=openapi_unified_path, levels=1)



from __future__ import annotations

__all__ = [
    "ApiName",
    "ApiEntityAlias",
    "UnifiedApiEntity",
    "ApiAliasToUnifiedEntities",
    "NamespaceInventory",
    "EntityURI",
]

ApiName = str
"""
The name of an API (which is also the stem of the API schema file itself).
"""


class ApiEntityAlias(str):
    """An API entity (typically an alias of the form "Tfl", "Tfl-1", "Tfl-2", etc.) from a
    non-unified API.
    """


class UnifiedApiEntity(str):
    """An API entity from the unified API"""


ApiAliasToUnifiedEntities = dict[ApiEntityAlias, list[UnifiedApiEntity]]
"""
An inventory of API entity aliases and the properly named entities they correspond to
from the unified API. Since there is no guarantee that different unified API entities
cannot have identical schemas, there may be a one-to-many mapping (i.e. the dict is
list-valued).
"""

NamespaceInventory = dict[ApiName, ApiAliasToUnifiedEntities]
"""
The namespace of all APIs and their inventories mapping each of their API entities
(aliases) to the properly named entities they correspond to from the unified API.
"""

EntityURI = str



"""Functions for the unique resolution of schema identifiers."""
from ...utils.lcp_trie import Trie
from ...utils.paths import SchemaPath

__all__ = [
    "find_backrefs",
    "find_string_literals",
    "find_array_literals",
    "find_chased",
    "find_arrays",
]


def find_backrefs(monoschema: dict) -> dict[str, SchemaPath]:
    """Map names of all schemas in the lookup to their referent (i.e. their array type)"""
    return {
        k: SchemaPath(v)
        for k, v in monoschema.items()
        if "items" in v
        if "$ref" in v["items"]
    }


def find_string_literals(monoschema: dict) -> dict[str, str]:
    """Map names of all schemas in the lookup with a common literal string."""
    return {k: v["type"] for k, v in monoschema.items() if v.get("type") == "string"}


def find_array_literals(monoschema: dict) -> dict[str, SchemaPath]:
    """Map names of all schemas in the lookup with a common literal array."""
    return {
        k: v["items"]["type"]
        for k, v in monoschema.items()
        if "type" in v.get("items", {})
    }


def find_chased(backrefs: dict[str, SchemaPath], name: str, ref_name: str) -> list[str]:
    return [
        backref
        for backref, schema_path in backrefs.items()
        if schema_path.ref.name == ref_name
    ]


def find_arrays(chased: list[str]) -> list[str]:
    """Can't simply choose the one(s) ending in 'Array': in Mode it ends in "Array-4" """
    t = Trie()
    for c in chased:
        t.insert(c)
    # array_shortlist = [c for c in chased if c.endswith("Array")]
    array_shortlist = []
    try:
        for stem, stem_node in t.root.data.items():
            if stem_node.is_word:
                array_shortlist.append(stem)
            else:
                substem, substem_node = next(iter(stem_node.data.items()))
                assert (
                    substem_node.is_word
                ), "Descended 2 trie levels and didn't get a word"
                array_shortlist.append(f"{stem}{substem}")
    except AssertionError:
        # It's an assortment of responses, just pull out the ones we want
        array_shortlist = [c for c in chased if "ApplicationJson" in c]
    return array_shortlist



import tubeulator
from pytest import mark


@mark.skiponci
def test_first_mode_is_bus():
    modes = tubeulator.fetch.line.meta_modes()
    first_mode_name = modes[0].ModeName
    assert first_mode_name == "bus"



from enum import Enum
from typing import TypeVar

__all__ = ["RouteEnum", "AnyEndpointRouteEnum"]


class RouteEnum(Enum):
    """Base route enum which can have methods set on (but no enum members)."""

    @property
    def parts(self):
        return self.value.split("/")


AnyEndpointRouteEnum = TypeVar("AnyEndpointRouteEnum", bound=RouteEnum)
"""Type indicating a specific endpoint (subclass of RouteEnum)"""



from enum import Enum

from .accident import AccidentStatsEndpointRoutes
from .air import AirQualityEndpointRoutes
from .bike import BikePointEndpointRoutes
from .crowding import CrowdingEndpointRoutes
from .disruption_lifts import DisruptionsLiftsEndpointRoutes
from .journey import JourneyEndpointRoutes
from .line import LineEndpointRoutes
from .mode import ModeEndpointRoutes
from .occupancy import OccupancyEndpointRoutes
from .place import PlaceEndpointRoutes
from .road import RoadEndpointRoutes
from .search import SearchEndpointRoutes
from .stop import StopPointEndpointRoutes
from .vehicle import VehicleEndpointRoutes

__all__ = ["EndpointRoute"]


class EndpointRoute(Enum):
    ACCIDENT_STATS = AccidentStatsEndpointRoutes
    AIR_QUALITY = AirQualityEndpointRoutes
    BIKE_POINT = BikePointEndpointRoutes
    CROWDING = CrowdingEndpointRoutes
    DISRUPTIONS_LIFTS_V2 = DisruptionsLiftsEndpointRoutes
    JOURNEY = JourneyEndpointRoutes
    LINE = LineEndpointRoutes
    MODE = ModeEndpointRoutes
    OCCUPANCY = OccupancyEndpointRoutes
    PLACE = PlaceEndpointRoutes
    ROAD = RoadEndpointRoutes
    SEARCH = SearchEndpointRoutes
    STOP_POINT = StopPointEndpointRoutes
    VEHICLE = VehicleEndpointRoutes



from .types import RouteEnum


class OccupancyEndpointRoutes(RouteEnum):
    BIKE_POINTS = "/BikePoints/{ids}"
    """Get the occupancy for bike points."""
    CAR_PARK = "/CarPark"
    """Gets the occupancy for all car parks that have occupancy data"""
    CAR_PARK_ID = "/CarPark/{id}"
    """Gets the occupancy for a car park with a given id"""
    CHARGE_CONNECTOR = "/ChargeConnector"
    """Gets the occupancy for all charge connectors"""
    CHARGE_CONNECTOR_IDS = "/ChargeConnector/{ids}"
    """Gets the occupancy for a charge connectors with a given id (sourceSystemPlaceId)"""



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from urllib.parse import urlencode

import httpx
from pydantic import BaseModel, TypeAdapter

from ..codegen import load_test
from ..db.store_creds import check_creds
from ..exc import RequestError
from ..utils.paths import (
    RefPath,
    SchemaPath,
    load_endpoint_component_schemas,
    load_endpoint_schema,
)
from .endpoint.names import EndpointNames
from .endpoint.routes.types import AnyEndpointRouteEnum

__all__ = ["TflApiPath", "Path", "Request", "GET"]


class TflApiPath:
    root: str = "https://api.tfl.gov.uk/"


@dataclass
class Path(TflApiPath):
    endpoint: EndpointNames
    route: AnyEndpointRouteEnum
    vars: list = field(init=False)

    class Var(Enum):
        DAY_OF_WEEK = "DayOfWeek"
        NAPTAN = "Naptan"
        DIRECTION = "direction"
        DISRUPTION_IDS = "disruptionIds"
        END_DATE = "endDate"
        FROM = "from"
        FROM_STOP_POINT_ID = "fromStopPointId"
        ID = "id"
        IDS = "ids"
        LAT = "lat"
        LINE = "line"
        LINE_ID = "lineId"
        LON = "lon"
        MODE = "mode"
        MODES = "modes"
        PAGE = "page"
        QUERY = "query"
        SEVERITY = "severity"
        START_DATE = "startDate"
        STOP_POINT_ID = "stopPointId"
        TO = "to"
        TO_STOP_POINT_ID = "toStopPointId"
        TYPE = "type"
        TYPES = "types"
        YEAR = "year"

    def __post_init__(self):
        self.vars = [
            self.Var(var.strip("{}"))
            for var in self.route.parts
            if var.startswith("{") and var.endswith("}")
        ]

    @property
    def url(self) -> str:
        return f"{self.root}{self.endpoint.value}{self.route.value}"

    @property
    def vars_to_fill(self) -> list[str]:
        return [var.value for var in self.vars]

    def check_vars_supplied(self, *args, **kwargs) -> dict:
        """Basic input validation, not type validation."""
        if args and kwargs:
            self.invalidate(
                "You can't mix args and kwargs, pick one.",
                rcvd={"args": args, "kwargs": kwargs},
            )
        elif self.vars and not (args or kwargs):
            self.invalidate("No vars supplied.", rcvd={"args": args, "kwargs": kwargs})
        if args and len(args) < len(self.vars):
            self.invalidate("Missing arguments", rcvd=args)
        elif kwargs and set(self.vars_to_fill) != set(kwargs):
            self.invalidate("Missing parameters", rcvd=kwargs)

    def invalidate(self, reason: str, rcvd: list[str] | dict):
        raise ValueError(f"Bad request ({reason}): {self.vars_to_fill=}, got {rcvd}")

    def build_url(self, *args, **kwargs) -> str:
        self.check_vars_supplied(*args, **kwargs)
        params = dict(zip(self.vars_to_fill, args)) if args else kwargs
        return self.url.format(**params)


@dataclass
class Request:
    """`tubeulator` uses :cls:`Request` objects to handle the various Transport for London
    APIs. They all follow the same structure and are therefore handled by one class.

    For the documentation on each particular API, look up the capitalised method name
    in the collected documentation below (which is just a shortened version of the
    Enums found in the `tubeulator.api.endpoint.routes` modules). For example the
    `tubeulator.fetch.stop_point.meta_modes` API is under `STOP_POINT` > `META_MODES`.

    EndpointRoute
    =============

    ACCIDENT_STATS
    AIR_QUALITY
    BIKE_POINT
    CROWDING
    DISRUPTIONS_LIFTS_V2
    JOURNEY
    LINE
    MODE
    OCCUPANCY
    PLACE
    ROAD
    SEARCH
    STOP_POINT
    VEHICLE

    ACCIDENT_STATS
    ==============

    YEAR_ACCIDENTS = "/{year}"
    Gets all accident details for accidents occuring in the specified year

    AIR_QUALITY
    ===========

    AIR_QUALITY_DATA = "/"
    Gets air quality data feed

    BIKE_POINT
    ==========

    ALL_LOCATIONS = "/"
    Gets all bike point locations. The Place object has an addtionalProperties array
    which contains the nbBikes, nbDocks and nbSpaces numbers which give the status of
    the BikePoint. A mismatch in these numbers i.e. nbDocks - (nbBikes + nbSpaces) != 0
    indicates broken docks.

    SEARCH_STATIONS = "/Search"
    Search for bike stations by their name, a bike point's name often contains
    information about the name of the street or nearby landmarks, for example. Note that the
    search result does not contain the PlaceProperties i.e. the status or occupancy of the
    BikePoint, to get that information you should retrieve the BikePoint by its id on
    '/BikePoint/id.

    BIKE_POINT_BY_ID = "/{id}"
    Gets the bike point with the given id.

    CROWDING
    ========

    NAPTAN_CROWDING = "/{Naptan}"
    Returns crowding information for Naptan

    LIVE_CROWDING = "/{Naptan}/Live"
    Returns live crowding information for Naptan

    DOW_CROWDING = "/{Naptan}/{DayOfWeek}"
    Returns crowding information for Naptan for Day of Week

    DISRUPTIONS_LIFTS_V2
    ====================

    ALL = "/"
    List of all currently disrupted lift routes

    JOURNEY
    =======

    FORWARD_REQUESTS = "/*"
    Forwards any remaining requests to the back-end

    JOURNEY_RESULTS = "/JourneyResults/{from}/to/{to}"
    Perform a Journey Planner search from the parameters specified in simple types

    MODES = "/Meta/Modes"
    Gets a list of all of the available journey planner modes

    LINE
    ====

    FORWARD_REQUESTS = "/*"
    Forwards any remaining requests to the back-end

    META_DISRUPTION_CATEGORIES = "/Meta/DisruptionCategories"
    Gets a list of valid disruption categories

    META_MODES = "/Meta/Modes"
    Gets a list of valid modes

    META_SERVICE_TYPES = "/Meta/ServiceTypes"
    Gets a list of valid ServiceTypes to filter on

    META_SEVERITY = "/Meta/Severity"
    Gets a list of valid severity codes

    LINES_BY_MODES = "/Mode/{modes}"
    Gets lines that serve the given modes.

    DISRUPTION_BY_MODES = "/Mode/{modes}/Disruption"
    Get disruptions for all lines of the given modes.

    ROUTE_BY_MODES = "/Mode/{modes}/Route"
    Gets all lines and their valid routes for given modes, including the name and id of
    the originating and terminating stops for each route

    STATUS_BY_MODES = "/Mode/{modes}/Status"
    Gets the line status of for all lines for the given modes

    ALL_ROUTES = "/Route"
    Get all valid routes for all lines, including the name and id of the originating and
    terminating stops for each route.

    SEARCH_LINES_ROUTES = "/Search/{query}"
    Search for lines or routes matching the query string

    STATUS_BY_SEVERITY = "/Status/{severity}"
    Gets the line status for all lines with a given severity. A list of valid severity
    codes can be obtained from a call to Line/Meta/Severity

    LINES_BY_IDS = "/{ids}"
    Gets lines that match the specified line ids.

    ARRIVALS_BY_IDS = "/{ids}/Arrivals"
    Get the list of arrival predictions for given line ids based at the given stop

    ARRIVALS_BY_IDS_STOP = "/{ids}/Arrivals/{stopPointId}"
    Get the list of arrival predictions for given line ids based at the given stop

    DISRUPTION_BY_IDS = "/{ids}/Disruption"
    Get disruptions for the given line ids

    ROUTE_BY_IDS = "/{ids}/Route"
    Get all valid routes for given line ids, including the name and id of the
    originating and terminating stops for each route.

    STATUS_BY_IDS = "/{ids}/Status"
    Gets the line status of for given line ids e.g Minor Delays

    STATUS_BY_IDS_DATES = "/{ids}/Status/{startDate}/to/{endDate}"
    Gets the line status for given line ids during the provided dates e.g Minor Delays

    ROUTE_SEQUENCE_BY_ID_DIRECTION = "/{id}/Route/Sequence/{direction}"
    Gets all valid routes for given line id, including the sequence of stops on each
    route.

    STOP_POINTS_BY_ID = "/{id}/StopPoints"
    Gets a list of the stations that serve the given line id

    TIMETABLE_BY_ID_FROM_STOP = "/{id}/Timetable/{fromStopPointId}"
    Gets the timetable for a specified station on the give line

    TIMETABLE_BY_ID_FROM_TO_STOP = (
        "/{id}/Timetable/{fromStopPointId}/to/{toStopPointId}"
    )
    Gets the timetable for a specified station on the give line with specified
    destination


    MODE
    ====

    ACTIVE_SERVICE_TYPES = "/ActiveServiceTypes"
    Returns the service type active for a mode. Currently only supports tube

    MODE_ARRIVALS = "/{mode}/Arrivals"
    Gets the next arrival predictions for all stops of a given mode

    OCCUPANCY
    =========

    BIKE_POINTS = "/BikePoints/{ids}"
    Get the occupancy for bike points.

    CAR_PARK = "/CarPark"
    Gets the occupancy for all car parks that have occupancy data

    CAR_PARK_ID = "/CarPark/{id}"
    Gets the occupancy for a car park with a given id

    CHARGE_CONNECTOR = "/ChargeConnector"
    Gets the occupancy for all charge connectors

    CHARGE_CONNECTOR_IDS = "/ChargeConnector/{ids}"
    Gets the occupancy for a charge connectors with a given id (sourceSystemPlaceId)

    PLACE
    =====

    PLACES_BY_GEO_REGION = "/"
    Gets the places that lie within a geographic region. The geographic region of interest can either
    be specified by using a lat/lon geo-point and a radius in metres to return places within the locus
    defined by the lat/lon of its centre or alternatively, by the use of a bounding box defined by the
    lat/lon of its north-west and south-east corners. Optionally filters on type and can strip
    properties for a smaller payload.

    FORWARD_REQUESTS = "/*"
    Forwards any remaining requests to the back-end

    META_CATEGORIES = "/Meta/Categories"
    Gets a list of all of the available place property categories and keys.

    META_PLACE_TYPES = "/Meta/PlaceTypes"
    Gets a list of the available types of Place.

    SEARCH_PLACES = "/Search"
    Gets all places that matches the given query

    PLACES_BY_TYPE = "/Type/{types}"
    Gets all places of a given type

    PLACE_BY_ID = "/{id}"
    Gets the place with the given id.

    PLACE_BY_TYPE_AT_COORDINATES = "/{type}/At/{lat}/{lon}"
    Gets any places of the given type whose geography intersects the given latitude and longitude. In
    practice this means the Place must be polygonal e.g. a BoroughBoundary.

    ROAD
    ====

    ALL_ROADS = "/"
    Gets all roads managed by TfL

    META_CATEGORIES = "/Meta/Categories"
    Gets a list of valid RoadDisruption categories

    META_SEVERITIES = "/Meta/Severities"
    Gets a list of valid RoadDisruption severity codes

    DISRUPTION_BY_IDS = "/all/Disruption/{disruptionIds}"
    Gets a list of active disruptions filtered by disruption Ids.

    STREET_DISRUPTION = "/all/Street/Disruption"
    Gets a list of disrupted streets. If no date filters are provided, current
    disruptions are returned.

    ROAD_BY_IDS = "/{ids}"
    Gets the road with the specified id (e.g. A1)

    ROAD_DISRUPTION = "/{ids}/Disruption"
    Get active disruptions, filtered by road ids

    ROAD_STATUS = "/{ids}/Status"
    Gets the specified roads with the status aggregated over the date range
    specified, or now until the end of today if no dates are passed.

    SEARCH
    ======

    SEARCH = "/"
    Search the site for occurrences of the query string. The maximum number of
    results returned is equal to the maximum page size of 100. To return subsequent
    pages, use the paginated overload.

    BUS_SCHEDULES = "/BusSchedules"
    Searches the bus schedules folder on S3 for a given bus number.

    META_CATEGORIES = "/Meta/Categories"
    Gets the available search categories.

    META_SEARCH_PROVIDERS = "/Meta/SearchProviders"
    Gets the available searchProvider names.

    META_SORTS = "/Meta/Sorts"
    Gets the available sorting options.

    STOP_POINT
    ==========

    WITHIN_RADIUS = "/"
    Gets a list of StopPoints within {radius} by the specified criteria

    FORWARD_REQUESTS = "/*"
    Forwards any remaining requests to the back-end

    META_CATEGORIES = "/Meta/Categories"
    Gets the list of available StopPoint additional information categories

    META_MODES = "/Meta/Modes"
    Gets the list of available StopPoint modes

    META_STOP_TYPES = "/Meta/StopTypes"
    Gets the list of available StopPoint types

    MODE = "/Mode/{modes}"
    Gets a list of StopPoints filtered by the modes available at that StopPoint.

    MODE_DISRUPTION = "/Mode/{modes}/Disruption"
    Gets a distinct list of disrupted stop points for the given modes

    SEARCH = "/Search"
    Search StopPoints by their common name, or their 5-digit Countdown Bus Stop Code.

    SEARCH_QUERY = "/Search/{query}"
    Search StopPoints by their common name, or their 5-digit Countdown Bus Stop Code.

    SERVICE_TYPES = "/ServiceTypes"
    Gets the service types for a given stoppoint

    SMS_ID = "/Sms/{id}"
    Gets a StopPoint for a given sms code.

    TYPE = "/Type/{types}"
    Gets all stop points of a given type

    TYPE_PAGE = "/Type/{types}/page/{page}"
    Gets all the stop points of given type(s) with a page number

    STOP_POINT_IDS = "/{ids}"
    Gets a list of StopPoints corresponding to the given list of stop ids.

    STOP_POINT_DISRUPTION = "/{ids}/Disruption"
    Gets all disruptions for the specified StopPointId, plus disruptions for any
    child Naptan records it may have.

    ARRIVAL_DEPARTURES = "/{id}/ArrivalDepartures"
    Gets the list of arrival and departure predictions for the given stop point id
    (overground and tfl rail only)

    ARRIVALS = "/{id}/Arrivals"
    Gets the list of arrival predictions for the given stop point id

    CAN_REACH_ON_LINE = "/{id}/CanReachOnLine/{lineId}"
    Gets Stopoints that are reachable from a station/line combination.

    CROWDING = "/{id}/Crowding/{line}"
    Gets all the Crowding data (static) for the StopPointId, plus crowding data
    for a given line and optionally a particular direction.

    DIRECTION_TO = "/{id}/DirectionTo/{toStopPointId}"
    Returns the canonical direction, "inbound" or "outbound", for a given pair of
    stop point Ids in the direction from -&gt; to.

    ROUTE = "/{id}/Route"
    Returns the route sections for all the lines that service the given stop point
    ids

    PLACE_TYPES = "/{id}/placeTypes"
    Get a list of places corresponding to a given id and place types.

    CAR_PARKS = "/{stopPointId}/CarParks"
    Get car parks corresponding to the given stop point id.

    TAXI_RANKS = "/{stopPointId}/TaxiRanks"
    Gets a list of taxi ranks corresponding to the given stop point id.

    VEHICLE
    =======

    EMISSION_SURCHARGE = "/EmissionSurcharge"
    Gets the Emissions Surcharge compliance for the Vehicle

    ULEZ_COMPLIANCE = "/UlezCompliance"
    Gets the Ulez Surcharge compliance for the Vehicle

    VEHICLE_ARRIVALS = "/{ids}/Arrivals"
    Gets the predictions for a given list of vehicle Id's.
    """

    path: Path

    def __call__(self, *args, **kwargs):
        try:
            result = self.send(*args, **kwargs)
        except httpx.HTTPError as exc:
            resp = exc.response
            raise RequestError(
                response=resp,
                path=self.path,
                _args=args,
                _kwargs=kwargs,
            ) from None
        parsed = self.parse(result)
        return parsed

    def send(self, *args, **kwargs) -> httpx.Response:
        url = self.path.build_url(*args, **kwargs)
        response = GET(url=url)
        response.raise_for_status()
        return response

    def ep_name(self, dehyphenate: bool = False) -> str:
        endpoint_name = self.path.endpoint.value
        return endpoint_name.replace("-", "") if dehyphenate else endpoint_name

    def response_refpath(self) -> RefPath:
        """Go from the endpoint route info to the response type (its reference type)."""
        endpoint_schema = load_endpoint_schema(self.ep_name())
        route_info = endpoint_schema["paths"][self.path.route.value]
        response_schema = route_info["get"]["responses"]["200"]["content"][
            "application/json"
        ]["schema"]
        response_ref = response_schema["$ref"]
        return RefPath(response_ref)

    def parse(self, response: httpx.Response):
        response_refpath = self.response_refpath()
        component_schemas = load_endpoint_component_schemas(self.ep_name())
        response_component_schema = component_schemas[response_refpath.name]
        if response_component_schema.get("items", {}).get("type") == "string":
            parsed = response.json()
        else:
            try:
                # Take a 2nd order reference
                ref_type = SchemaPath(response_component_schema)
                ref_name = ref_type.ref.name
            except Exception:
                # Not a 2nd order reference
                ref_name = response_refpath.name
            try:
                marshals = getattr(
                    load_test,
                    self.ep_name(dehyphenate=True),
                ).Deserialisers
            except Exception as exc:
                hint = "(did the import from `generated` get removed by a linter?)"
                fail_msg = f"The API endpoint wasn't attached to `load_test` ({hint})"
                raise RuntimeError(f"{fail_msg} -- {exc}")
            dto = marshals.select_component(ref_name).value
            is_pydantic = issubclass(dto, BaseModel)
            try:
                if is_pydantic:
                    # Peek at the first character, let Pydantic deserialise the JSON
                    is_array = response.text[0] == "["
                    # data = response.json()
                    # is_array = isinstance(data, list)
                    if is_array:
                        validate = TypeAdapter(list[dto]).validate_json
                    else:
                        validate = dto.model_validate_json
                    parsed = validate(response.content)
                else:
                    # Assume JSONWizard
                    parsed = dto.from_json(response.content)
            except:
                raise
        return parsed


def GET(url: str) -> httpx.Response:
    credentials = check_creds()
    params = urlencode(
        {"app_id": credentials["app_id"], "app_key": credentials["primary_key"]},
    )
    return httpx.get(url, params=params)



"""Helpers courtesy of the dataclass_wizard package's `utils.string_conv` module.

These are used to convert the keys from Pydantic model field names to TfL Data Transfer
Object (DTO) JSON keys.

Pydantic uses 'alias generators' for this, the same as DCW just in the opposite
direction (DCW consider conversions going from JSON key to model field). Unfortunately
Pydantic's string conversion helpers do not actually do the job required here,
so I copied over these helper functions rather than enforce a dependency for them.
"""
import re

__all__ = ["to_camel_case", "to_pascal_case"]


def replace_multi_with_single(string: str, char="_") -> str:
    """
    Replace multiple consecutive occurrences of `char` with a single one.
    """
    rep = char + char
    while rep in string:
        string = string.replace(rep, char)

    return string


def to_camel_case(string: str) -> str:
    """
    Convert a string to Camel Case.

    Examples::

        >>> to_camel_case("ModeName")
        'modeName'
        >>> to_camel_case("a_b_c")
        'aBC'

    """
    string = replace_multi_with_single(string.replace("-", "_").replace(" ", "_"))

    return string[0].lower() + re.sub(
        r"(?:_)(.)",
        lambda m: m.group(1).upper(),
        string[1:],
    )


def to_pascal_case(string):
    """
    Converts a string to Pascal Case (also known as "Upper Camel Case")

    Examples::

        >>> to_pascal_case("device_type")
        'DeviceType'

    """
    string = replace_multi_with_single(string.replace("-", "_").replace(" ", "_"))

    return string[0].upper() + re.sub(
        r"(?:_)(.)",
        lambda m: m.group(1).upper(),
        string[1:],
    )
